package com.streambase.contrib.erjang.impl;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseRuntimeException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;

import erjang.EAtom;
import erjang.EObject;
import erjang.ETuple;
import erjang.ETuple2;
import erjang.RPC;

public class SimpleEmbedded {
	private static final ErjangSession SESSION = new ErjangSession();
	
	private static final EAtom am_callerl = EAtom.intern("callerl");
	private static final EAtom am_init = EAtom.intern("init");
	private static final EAtom am_accumulate = EAtom.intern("accumulate");
	private static final EAtom am_emit = EAtom.intern("emit");

	private SimpleEmbedded() {
		// Monostate
	}

	public static EObject callerl_init(String mod, EObject state) {
		EAtom m = EAtom.intern(mod);
		ETuple et =  (ETuple)RPC.call(m, am_init, state);
		return et.elm(2); // Unwrap Erjang RPC call response
	}

	public static EObject callerl_accumulate(String mod, EObject state, Tuple args, Tuple hint) {
		EAtom m = EAtom.intern(mod);
		ETuple et = (ETuple)RPC.call(
			m, am_accumulate, state, 
			TypeMapping.sbToErjang(args, CompleteDataType.forTuple(args.getSchema())));
		return et.elm(2);
	}
	
	public static Tuple callerl_emit(String mod, EObject state, Tuple hint) {
		EAtom m = EAtom.intern(mod);
		EObject r = RPC.call(m, am_emit, state);
		try {
			ETuple2 t = (ETuple2)r;
			if (!t.elm(1).equals(EAtom.intern("ok"))) {
				// @NOTE: Response could be an error tuple - TBD!
				throw new StreamBaseRuntimeException("Feck: " + r.toString());
			}
			TypeMapping.BestGuess bg = TypeMapping.erjangToCdt(t.elm(2)); // SB and Erlang type systems significantly different
			
			return wrapTuple(bg);
		} catch (TupleException e) {
			throw new StreamBaseRuntimeException(e);
		}
	}

	static Tuple wrapTuple(TypeMapping.BestGuess bg) throws TupleException {
		// Wrap response as a StreamBase 'tuple'
		Schema TMP = new Schema("TmpResult", new Schema.Field[] { new Schema.Field("Result", bg.guess)});
		Tuple tmp = TMP.createTuple();
		tmp.setField("Result", bg.sbType);
		return tmp;
	}

	public static Tuple callerl(String mod, Tuple args) {
		EAtom m = EAtom.intern(mod);
		EObject r = RPC.call(m, am_callerl, new EObject[] { TypeMapping.sbToErjang(args, CompleteDataType.forTuple(args.getSchema()))});
		try {
			ETuple2 t = (ETuple2)r;
			if (!t.elm(1).equals(EAtom.intern("ok"))) {
				throw new StreamBaseRuntimeException("Feck");
			}
			TypeMapping.BestGuess bg = TypeMapping.erjangToCdt(t.elm(2));
			Tuple tmp = wrapTuple(bg);
			return tmp;
		} catch (TupleException e) {
			throw new StreamBaseRuntimeException(e);
		}
	}

}
