package com.streambase.contrib.erjang.impl;

import java.sql.Date;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import com.streambase.sb.ByteArrayView;
import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.NullValueException;
import com.streambase.sb.Schema;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;

import erjang.EAtom;
import erjang.EBig;
import erjang.EBinary;
import erjang.EDouble;
import erjang.EObject;
import erjang.ERT;
import erjang.ESeq;
import erjang.ESmall;
import erjang.EString;
import erjang.ETuple;

public class TypeMapping {
	public static class BestGuess {
		CompleteDataType guess;
		Object sbType;
		public BestGuess(CompleteDataType c, Object o) {
			guess = c;
			sbType = o;
		}
	}

	private TypeMapping() {
		// Monostate
	}

	public static EObject sbToErjang(Tuple t, Schema.Field f) throws NullValueException, TupleException {
		CompleteDataType c = f.getCompleteDataType();
		if (CompleteDataType.isSimpleType(c.getDataType())) {
			return TypeMapping.sbToErjang(t.getField(f.getName()), c);
		}
		if (DataType.TUPLE.equals(c.getDataType())) {
			if (t.isNull(f)) return ERT.NIL;
			return TypeMapping.sbToErjang(t.getField(f.getName()),c);
		}
		if (DataType.LIST.equals(c.getDataType())) {
			if (t.isNull(f)) return ERT.NIL;
			return TypeMapping.sbToErjang(t.getList(f),c.getElementType());
		}
		throw new UnsupportedOperationException();
	}

	public static EObject sbToErjang(Schema s, Tuple t) {
		if (t == null) return ERT.NIL;
		EObject[] fs = new EObject[s.getFieldCount()];
		int count = 0;
		for (Schema.Field nf : t.getSchema().getFields()) {
			fs[count++] = TypeMapping.sbToErjang(t.getField(nf), nf.getCompleteDataType());
		}
		return ETuple.make(fs);
	}

	public static EObject sbToErjang(List<Object> l, CompleteDataType c) {
		List<EObject> r = new ArrayList<EObject>();
		for(Object o : l) {
			r.add(TypeMapping.sbToErjang(o, c));
		}
		return ESeq.fromList(r);
	}

	@SuppressWarnings("unchecked")
	static EObject sbToErjang(Object o, CompleteDataType c) {
		if (o instanceof List) {
			return sbToErjang((List<Object>)o, c);
		}
		if (o instanceof Tuple) {
			return sbToErjang(((Tuple)o).getSchema(), (Tuple)o);
		}
		if (CompleteDataType.forBlob().equals(c)) {
			return o == null ? ERT.NIL : new EBinary(((ByteArrayView)o).array());
		}
		if (CompleteDataType.forBool().equals(c)) {
			return o == null ? ERT.NIL : EAtom.intern((Boolean)o ? "true" : "false");
		}
		if (CompleteDataType.forDouble().equals(c)) {
			return o == null ? ERT.NIL : new EDouble((Double)o); 
		}
		if (CompleteDataType.forLong().equals(c)) {
			return o == null ? ERT.NIL : new EBig((Long)o);
		}
		if (CompleteDataType.forInt().equals(c)) {
			return o == null ? ERT.NIL : new EBig((Integer)o);
		}
		if (CompleteDataType.forString().equals(c)) {
			return o == null ? ERT.NIL : new EString((String)o);
		}
		if (CompleteDataType.forTimestamp().equals(c)) {
			if (o == null) return ERT.NIL;
			Timestamp ts = (Timestamp)o;
			if (Timestamp.INTERVAL == ts.getType()) {
				return ETuple.make(EAtom.intern("interval"), new EBig(ts.toMsecs()));
			} else {
				long tsx = ts.toMsecs();
				long mega = tsx/1000000000;
				long sec = (tsx%1000000000)/1000;
				long micros = (tsx%1000)*1000;
				return ETuple.make(EAtom.intern("timestamp"), ETuple.make(new EBig(mega), new EBig(sec), new EBig(micros)));
			}
		}
		throw new UnsupportedOperationException();
	}

	public static Tuple erjangToTuple(Tuple t, Schema.Field f, EObject o) throws TupleException {
		if (o.testNil() == ERT.NIL) {
			t.setNull(f);
			return t;
		}
		if (CompleteDataType.isSimpleType(f.getDataType())) {
			t.setField(f,TypeMapping.erjangToSbType(o, f.getCompleteDataType()));
			return t;
		}
		if (DataType.LIST.equals(f.getDataType())) {
			t.setField(f,TypeMapping.erjangToSbType(o, f.getElementType()));
			return t;
		}
		if (DataType.TUPLE.equals(f.getDataType())) {
			t.setField(f,TypeMapping.erjangToSbType(o, CompleteDataType.forTuple(f.getCompleteDataType().getSchema())));
			return t;
		}
		throw new UnsupportedOperationException();
	}

	static Object erjangToSbType(EObject o, CompleteDataType c) throws TupleException {
		if (o.testNil() == ERT.NIL) {
			return ERT.NIL;
		}
		EAtom a = o.testAtom();
		if (a != null) {
			if ("false".equals(a.toString())) {
				return Boolean.FALSE;
			}
			if ("true".equals(a.toString())) {
				return Boolean.TRUE;
			}
		}
		EBig b = o.testBig();
		if (b != null) {
			if (CompleteDataType.forLong().equals(c)) {
				return b.longValue();
			} else {
				return b.intValue();
			}
		}
		ESmall is = o.testSmall();
		if (is != null) {
			return is.asInt();
		}
		EString s = o.testString();
		if (s != null && CompleteDataType.forString().equals(c)) {
			return s.stringValue();
		}
		final EBinary ebin = o.testBinary();
		if (ebin != null && CompleteDataType.forBlob().equals(c)) {
			return ByteArrayView.makeView(ebin.toByteArray());
		}
		EDouble d = o.testFloat();
		if (d != null) {
			return d.doubleValue();
		}
		ESeq l = o.testSeq();
		if (l != null) {
			List<Object> tl = new ArrayList<Object>();
			do {
				EObject x = l.head();
				tl.add(erjangToSbType(x,c));
				l = l.tail();
			} while(l != ERT.NIL);
			return tl;
		}
		ETuple x = o.testTuple();
		if (x != null) {
			if (EAtom.intern("interval").equals(x.elm(1))) {
				return Timestamp.msecs(Timestamp.INTERVAL,((EBig)x.elm(2)).longValue());
			}
			if (EAtom.intern("timestamp").equals(x.elm(1))) {
				ETuple ts = ((ETuple)x.elm(2));
				long mega =((EBig)ts.elm(1)).longValue();
				long secs = ((EBig)ts.elm(2)).longValue();
				long micros = ((EBig)ts.elm(3)).longValue();
				long tsx = ((mega*1000000000) + (secs*1000) + (micros/1000));
				return new Timestamp(new Date(tsx));
			}
			Tuple tt = c.getSchema().createTuple();
			for (int i = 0; tt != null && i < tt.getSchema().getFieldCount(); i++) {
				erjangToTuple(tt, tt.getSchema().getField(i), x.elm(i+1));
			}
			return tt;
		}
		throw new UnsupportedOperationException();
	}

	public static BestGuess erjangToCdt(EObject o) throws TupleException {
		if (o.testNil() == ERT.NIL) {
			throw new UnsupportedOperationException(); // no type information :(
		}
		if (o.testAtom() != null) {
			if ("false".equals(o.toString()) || "true".equals(o.toString())) {
				return new BestGuess(CompleteDataType.forBool(),false);
			}
			return new BestGuess(CompleteDataType.forString(),((EString)o).toString());
		}
		if (o.testString() != null) {
			return new BestGuess(CompleteDataType.forString(),((EString)o).toString());
		}
		EBig b = o.testBig();
		if (b != null) {
			return new BestGuess(CompleteDataType.forLong(), b.longValue());
		}
		ESmall is = o.testSmall();
		if (is != null) {
			return new BestGuess(CompleteDataType.forInt(), is.intValue());
		}
		final EBinary ebin = o.testBinary();
		if (ebin != null) {
			return new BestGuess(CompleteDataType.forBlob(),ByteArrayView.makeView(ebin.toByteArray()));
		}
		EDouble d = o.testFloat();
		if (d != null) {
			return new BestGuess(CompleteDataType.forDouble(),d.doubleValue());
		}
		ESeq l = o.testSeq();
		if (l != null) {
			List<Object> tl = new ArrayList<Object>();
			BestGuess bg = null;
			do {
				EObject x = l.head();
				bg = erjangToCdt(x);
				tl.add(bg.sbType);
				l = l.tail();
			} while(l != ERT.NIL);
			return new BestGuess(CompleteDataType.forList(bg.guess), tl);
		}
	
		ETuple x = o.testTuple();
		if (x != null) {
			if (EAtom.intern("interval").equals(x.elm(1)) || EAtom.intern("timestamp").equals(x.elm(1))) {
				if (EAtom.intern("interval").equals(x.elm(1))) {
					return new BestGuess(CompleteDataType.forTimestamp(),Timestamp.msecs(Timestamp.INTERVAL,((EBig)x.elm(2)).longValue()));
				}
				if (EAtom.intern("timestamp").equals(x.elm(1))) {
					ETuple ts = ((ETuple)x.elm(2));
					long mega =((EBig)ts.elm(1)).longValue();
					long secs = ((EBig)ts.elm(2)).longValue();
					long micros = ((EBig)ts.elm(3)).longValue();
					long tsx = ((mega*1000000000) + (secs*1000) + (micros/1000));
					return new BestGuess(CompleteDataType.forTimestamp(),new Timestamp(new Date(tsx)));
				}
			}
			List<Schema.Field> fields = new LinkedList<Schema.Field>();
			int count = 0;
			for (int i = 0; i < x.arity(); i++) {
				BestGuess bg = erjangToCdt(x.elm(i+1));
				fields.add(new Schema.Field("Erj" + count++, bg.guess));
			}
			Schema s = new Schema("FromErjangResultSchema", fields);
			Tuple t = s.createTuple();
			for (int i = 0; i < x.arity(); i++) {
				Object to = erjangToSbType(x.elm(i+1), s.getField(i).getCompleteDataType());
				t.setField(i,to);
			}
			return new BestGuess(CompleteDataType.forTuple(s), t);
		}
		throw new UnsupportedOperationException();
	}
}